#include "DXApplication.h"
#include "Timer.h"
#include "Logger.h"
#include "SceneManager.h"
#include "Model.h"
#include "Entity.h"
#include "Camera.h"

#ifndef HID_USAGE_PAGE_GENERIC
#define HID_USAGE_PAGE_GENERIC		((USHORT) 0x01)
#endif
#ifndef HID_USAGE_GENERIC_MOUSE
#define HID_USAGE_GENERIC_MOUSE		((USHORT) 0x02)
#endif

namespace MyEngine
{
	DXApplication* DXApplication::m_pInstance = NULL;

	DXApplication* DXApplication::Instance()
	{
		if(m_pInstance == NULL)
			m_pInstance = new DXApplication;

		return m_pInstance;
	}

	void DXApplication::DeleteInstance()
	{
		delete m_pInstance;
		m_pInstance = NULL;
	}

	DXApplication::DXApplication()
		: m_hInstance(NULL), m_hWnd(NULL),
		m_pd3dObject(NULL), m_pd3dDevice(NULL), m_pd3dFont(NULL), m_bDeviceLost(false),
		m_nWidth(800), m_nHeight(600), m_nRefreshRate(60),
		m_nFrameCount(0)
	{
		m_strWindowClass = _T("D3DWINDOWCLASS");
		m_strWindowTitle = _T("D3D Window");

		for(uint i=0; i < 256; ++i)
			m_aKeyboardState[i] = false;

		Logger::Instance()->Write(_T("Log created."));
	}

	DXApplication::~DXApplication()
	{
		ModelMap::iterator i = m_models.begin();

		while(i != m_models.end())
		{
			delete i->second;
			i->second = NULL;
			i++;
		}

		delete m_pSceneManager;

		freeResources();

		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pd3dObject);

		Timer::DeleteInstance();
		Logger::DeleteInstance();
	}

	bool DXApplication::Initialize(HINSTANCE hInstance)
	{
		m_hInstance = hInstance;

		readSettings();

		if(!initWindow())
			return false;

		if(!initInput())
			return false;

		if(!initD3D())
			return false;

		if(!initFont())
			return false;

		if(!loadModels())
			return false;

		if(!initScene())
			return false;

		return true;
	}

	void DXApplication::readSettings()
	{
		std::ifstream settingsFile("settings.ini");

		if (!settingsFile)
		{
			Logger::Instance()->Write(_T("Couldn't open settings file."));
		}
		else
		{
			settingsFile >> m_nWidth;
			settingsFile >> m_nHeight;
			settingsFile >> m_nRefreshRate;
		}
	}

	bool DXApplication::initWindow()
	{
		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(wcex);
		wcex.style = CS_CLASSDC;
		wcex.lpfnWndProc = WndProcWrapper;
		wcex.cbClsExtra = 0;
		wcex.cbWndExtra = 0;
		wcex.hInstance = m_hInstance;
		wcex.hIcon = NULL;
		wcex.hCursor = NULL;
		wcex.hbrBackground = NULL;
		wcex.lpszMenuName = NULL;
		wcex.lpszClassName = m_strWindowClass.c_str();
		wcex.hIconSm = NULL;

		if(RegisterClassEx(&wcex) == 0)
			return false;

		m_hWnd = CreateWindowEx(WS_EX_TOPMOST, m_strWindowClass.c_str(), m_strWindowTitle.c_str(),
			WS_POPUP /*WS_OVERLAPPEDWINDOW*/, 0, 0, m_nWidth, m_nHeight, NULL, NULL, m_hInstance, NULL);

		if(m_hWnd == NULL)
			return false;

		ShowWindow(m_hWnd, SW_SHOW);
		UpdateWindow(m_hWnd);

		return true;
	}

	bool DXApplication::initInput()
	{
		RAWINPUTDEVICE Rid[1];
		Rid[0].usUsagePage = HID_USAGE_PAGE_GENERIC;
		Rid[0].usUsage = HID_USAGE_GENERIC_MOUSE;
		Rid[0].dwFlags = RIDEV_INPUTSINK;
		Rid[0].hwndTarget = m_hWnd;

		if(::RegisterRawInputDevices(Rid, 1, sizeof(Rid[0])) == FALSE)
			return false;

		return true;
	}

	bool DXApplication::initD3D()
	{
		// Create Direct3D 9 object.
		m_pd3dObject = Direct3DCreate9(D3D_SDK_VERSION);
		if(m_pd3dObject == NULL)
		{
			Logger::Instance()->Write(_T("FAILED: Create Direct3D object."));
			return false;
		}

		D3DDISPLAYMODE dm;
		if(!findReqAdapterMode(D3DFMT_X8R8G8B8, &dm))
		{
			Logger::Instance()->Write(_T("FAILED: Couldn't find appropriate adapter mode."));
			return false;
		}

		// Fill in presentation parameters.
		ZeroMemory(&m_d3dPP, sizeof(m_d3dPP));

		/*
		m_d3dPP.Windowed = TRUE;
		m_d3dPP.SwapEffect = D3DSWAPEFFECT_DISCARD;
		m_d3dPP.BackBufferCount = 1;
		m_d3dPP.BackBufferFormat = D3DFMT_UNKNOWN;
		m_d3dPP.BackBufferWidth = 0;
		m_d3dPP.BackBufferHeight = 0;
		m_d3dPP.EnableAutoDepthStencil = TRUE;
		m_d3dPP.AutoDepthStencilFormat = D3DFMT_D16;
		*/

		m_d3dPP.Windowed = FALSE;
		m_d3dPP.SwapEffect = D3DSWAPEFFECT_DISCARD;
		m_d3dPP.BackBufferCount = 1;
		m_d3dPP.BackBufferFormat = D3DFMT_A8R8G8B8;
		m_d3dPP.BackBufferWidth = dm.Width;
		m_d3dPP.BackBufferHeight = dm.Height;
		m_d3dPP.EnableAutoDepthStencil = TRUE;
		m_d3dPP.AutoDepthStencilFormat = D3DFMT_D16;
		m_d3dPP.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
		m_d3dPP.FullScreen_RefreshRateInHz = dm.RefreshRate;

		// Create Direct3D 9 device.
		HRESULT hr = m_pd3dObject->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL,
			m_hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &m_d3dPP, &m_pd3dDevice);

		if(FAILED(hr))
		{
			Logger::Instance()->Write(_T("FAILED: Create Direct3D device."));
			return false;
		}

		m_nTimeStep = 1000 / m_nRefreshRate;
		m_nPrevFrameTime = Timer::Instance()->Milliseconds();
		m_nNextFrameTime = m_nPrevFrameTime + m_nTimeStep;

		return true;
	}

	bool DXApplication::findReqAdapterMode(D3DFORMAT fmt, D3DDISPLAYMODE* pDisplayMode)
	{
		bool bRequestedModeFound = false;

		uint nModeCount = m_pd3dObject->GetAdapterModeCount(D3DADAPTER_DEFAULT, fmt);

		Logger::Instance()->Write(_T("Listing available modes for default adapter..."));

		for(uint i=0; i < nModeCount; ++i)
		{
			D3DDISPLAYMODE dm;
			HRESULT hr = m_pd3dObject->EnumAdapterModes(D3DADAPTER_DEFAULT, fmt, i, &dm);

			if(FAILED(hr))
				return false;

			TOStringStream output;

			// Check whether this mode satisfies the requirements.
			if(!bRequestedModeFound &&
				dm.Format == fmt &&
				dm.Width == m_nWidth &&
				dm.Height == m_nHeight &&
				dm.RefreshRate == m_nRefreshRate)
			{
				*pDisplayMode = dm;
				bRequestedModeFound = true;
				output << "-> ";
			}
			else
			{
				output << "   ";
			}

			output << dm.Width << "x" << dm.Height << " @ " << dm.RefreshRate << "Hz";
			Logger::Instance()->Write(output.str());
		}

		return bRequestedModeFound;
	}

	bool DXApplication::initFont()
	{
		HDC hDC = GetDC(NULL);
		int nLogPixelsY = GetDeviceCaps(hDC, LOGPIXELSY);
		ReleaseDC(NULL, hDC);

		int nHeight = -14 * nLogPixelsY / 72;

		HRESULT hr;
		hr = D3DXCreateFont(m_pd3dDevice,	// D3D device
			nHeight,						// Height
			0,								// Width
			0/*FW_BOLD*/,					// Weight
			1,								// MipLevels, 0 = autogen mipmaps
			FALSE,							// Italic
			DEFAULT_CHARSET,				// CharSet
			OUT_DEFAULT_PRECIS,				// OutputPrecision
			DEFAULT_QUALITY,				// Quality
			DEFAULT_PITCH | FF_DONTCARE,	// PitchAndFamily
			_T("Arial"),					// pFaceName
			&m_pd3dFont);					// ppFont

		if(FAILED(hr))
		{
			Logger::Instance()->Write(_T("FAILED: Create Direct3D font."));
			return false;
		}

		return true;
	}

	// /////////////////////////////////////////////////////////////////
	// / Load a model from file and add it to model collection.
	// /////////////////////////////////////////////////////////////////
	bool DXApplication::loadModel(const TString& strFilename)
	{
		if(m_models.find(strFilename) != m_models.end())
			return true;

		Model* m = new Model(m_pd3dDevice);

		if(!m->Load(strFilename))
			return false;

		Logger::Instance()->Write(_T("Successfully loaded ") + strFilename);
		m_models[strFilename] = m;

		return true;
	}

	bool DXApplication::loadModels( )
	{
		if(!loadModel(_T("Models\\Crate\\Crate.x")))
			return false;

		if(!loadModel(_T("Models\\Crate\\Crate.x")))
			return false;

		if(!loadModel(_T("Models\\Level\\Level.x")))
			return false;

		if(!loadModel(_T("Models\\Sniper\\Sniper.x")))
			return false;

		if(!loadModel(_T("Models\\Skybox\\Skybox.x")))
			return false;

		return true;
	}

	Model* DXApplication::getModelById(const TString& strId)
	{
		ModelMap::iterator i = m_models.find(strId);

		if(i != m_models.end())
			return i->second;

		return NULL;
	}

	bool DXApplication::initScene()
	{
		m_pSceneManager = new SceneManager(m_pd3dDevice);

		// /////////////////////////////////////////////////////////////////
		// / Place entities.
		// /////////////////////////////////////////////////////////////////
		Entity* ent1 = new Entity;
		ent1->SetModel(getModelById(_T("Models\\Crate\\Crate.x")));
		ent1->SetPosition( D3DXVECTOR3( 0.0F, 25.0F, -250.0F ) );
		m_pSceneManager->AddEntity(ent1);
		m_pSceneManager->SetPlayer(ent1);

		Entity* ent2 = new Entity;
		ent2->SetModel(getModelById(_T("Models\\Level\\Level.x")));
		m_pSceneManager->AddEntity(ent2);

		Entity* ent3 = new Entity;
		ent3->SetModel(getModelById(_T("Models\\Skybox\\Skybox.x")));
		ent3->SetPosition( D3DXVECTOR3(0.0F, 25.0F, 0.0F) );
		m_pSceneManager->SetSkyBox(ent3);

		Entity* ent4 = new Entity;
		ent4->SetModel(getModelById(_T("Models\\Crate\\Crate.x")));
		ent4->SetPosition( D3DXVECTOR3(0.0F, 5.0F, -20.0F) );
		ent4->SetAngle( D3DXVECTOR3(D3DX_PI/3.0F, 0.0F, 0.0F) );
		m_pSceneManager->AddEntity(ent4);

		Entity* ent5 = new Entity;
		ent5->SetModel(getModelById(_T("Models\\Crate\\Crate.x")));
		ent5->SetPosition( D3DXVECTOR3(10.0F, 5.0F, 20.0F) );
		ent5->SetAngle( D3DXVECTOR3(D3DX_PI/5.0F, 0.0F, 0.0F) );
		m_pSceneManager->AddEntity(ent5);

		Entity* ent6 = new Entity;
		ent6->SetModel(getModelById(_T("Models\\Sniper\\Sniper.x")));
		ent6->SetPosition( D3DXVECTOR3(0.0F, 5.0F, 120.0F) );
		ent6->SetAngVelocity( D3DXVECTOR3(D3DX_PI/4.0F, 0.0F, 0.0F) );
		m_pSceneManager->AddEntity(ent6);

		// /////////////////////////////////////////////////////////////////
		// / Setup the camera.
		// /////////////////////////////////////////////////////////////////
		Camera* cam = new Camera;
		cam->AttachToEntity( ent1 );
		cam->SetAspectRatio((float) m_nWidth / m_nHeight);
		cam->SetPosition(D3DXVECTOR3(0.0F, 55.0F, 0.0F));
		//cam->SetPitch( D3DX_PI/4.0F );
		m_pSceneManager->SetCamera(cam);

		return true;
	}

	void DXApplication::processKeyboard()
	{
		// /////////////////////////////////////////////////////////////////
		// / Quit.
		// /////////////////////////////////////////////////////////////////
		if(isKeyDown(VK_ESCAPE))
		{
			Logger::Instance()->Write(_T("Terminating... bye!"));
			PostQuitMessage(0);
		}

		D3DXVECTOR3 axes[3];
		m_pSceneManager->Player()->OrientationAxes(axes[0], axes[1], axes[2]);

		D3DXVECTOR3 velocity(0.0F, 0.0F, 0.0F);

		// /////////////////////////////////////////////////////////////////
		// / Walk forward/backward.
		// /////////////////////////////////////////////////////////////////
		if(isKeyDown('W') || isKeyDown(VK_UP))
			velocity += axes[2];

		else if(isKeyDown('S') || isKeyDown(VK_DOWN))
			velocity -= axes[2];

		// /////////////////////////////////////////////////////////////////
		// / Strafe left/right.
		// /////////////////////////////////////////////////////////////////
		if(isKeyDown('A') || isKeyDown(VK_LEFT))
			velocity -= axes[0];

		else if(isKeyDown('D') || isKeyDown(VK_RIGHT))
			velocity += axes[0];

		// /////////////////////////////////////////////////////////////////
		// / Fly up/down.
		// /////////////////////////////////////////////////////////////////
		if(isKeyDown(VK_PRIOR))
			velocity += axes[1];

		else if(isKeyDown(VK_NEXT))
			velocity -= axes[1];

		// /////////////////////////////////////////////////////////////////
		// / Set magnitude of velocity.
		// /////////////////////////////////////////////////////////////////
		D3DXVec3Normalize(&velocity, &velocity);
		velocity *= 50.0F;

		m_pSceneManager->Player()->SetVelocity(velocity);
	}

	void DXApplication::freeResources()
	{
		SAFE_RELEASE(m_pd3dFont);
	}

	void DXApplication::restoreResources()
	{
		initFont();
	}

	void DXApplication::updateScene(uint ms)
	{
		m_pSceneManager->UpdateScene(ms);
	}

	void DXApplication::handleLostDevice()
	{
		Sleep(250);

		HRESULT hr = m_pd3dDevice->TestCooperativeLevel();

		if(SUCCEEDED(hr))
		{
			Logger::Instance()->Write(_T("Direct3D device restored."));
			m_bDeviceLost = false;
		}
		else if(hr == D3DERR_DEVICENOTRESET)
		{
			freeResources();

			if(FAILED(m_pd3dDevice->Reset(&m_d3dPP)))
				return;

			restoreResources();
		}
	}

	void DXApplication::renderScene()
	{
		if(m_bDeviceLost)
		{
			handleLostDevice();
			return;
		}

		m_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(128,128,128), 1.0F, 0);

		if(SUCCEEDED(m_pd3dDevice->BeginScene()))
		{
			m_pSceneManager->RenderScene();

			TString strEscape = _T("Press 'Esc' to quit.");
			TString strMove = _T("Press W, A, S, D, PgUp and PgDn to move.");

			TOStringStream mode;
			mode << m_nWidth << _T(" x ") << m_nHeight << _T(" @ ") << m_nRefreshRate << " Hz";

			TOStringStream fps;
			fps << _T("FPS: ") << m_fFps;

			RECT rc;

			SetRect(&rc, 10, 10, 200, 30);
			m_pd3dFont->DrawText(NULL, strEscape.c_str(), -1, &rc, DT_NOCLIP, D3DXCOLOR(1.0F,1.0F,0.0F,1.0F));

			SetRect(&rc, 10, 30, 200, 50);
			m_pd3dFont->DrawText(NULL, strMove.c_str(), -1, &rc, DT_NOCLIP, D3DXCOLOR(1.0F,1.0F,0.0F,1.0F));

			SetRect(&rc, 10, 55, 200, 75);
			m_pd3dFont->DrawText(NULL, mode.str().c_str(), -1, &rc, DT_NOCLIP, D3DXCOLOR(1.0F,1.0F,1.0F,1.0F));

			SetRect(&rc, 10, 75, 200, 95);
			m_pd3dFont->DrawText(NULL, fps.str().c_str(), -1, &rc, DT_NOCLIP, D3DXCOLOR(1.0F,1.0F,1.0F,1.0F));

			m_pd3dDevice->EndScene();
		}

		HRESULT hr = m_pd3dDevice->Present(NULL, NULL, NULL, NULL);

		if(hr == D3DERR_DEVICELOST)
		{
			Logger::Instance()->Write(_T("Direct3D device lost."));
			m_bDeviceLost = true;
		}
	}

	void DXApplication::makeStep()
	{
		uint nCurrentTime = Timer::Instance()->Milliseconds();

		// Calculate FPS
		float fFpsElapsedTime = (nCurrentTime - m_nFpsStartTime) / 1000.0F;
		if(fFpsElapsedTime > 1.0F)
		{
			m_fFps = m_nFrameCount / fFpsElapsedTime;
			m_nFpsStartTime = nCurrentTime;
			m_nFrameCount = 0;
		}

		// Update scene and render it
		if(nCurrentTime > m_nNextFrameTime)
		{
			++m_nFrameCount;

			processKeyboard();

			updateScene(nCurrentTime - m_nPrevFrameTime);
			renderScene();

			m_nPrevFrameTime = nCurrentTime;
			m_nNextFrameTime += m_nTimeStep;

			if(nCurrentTime > m_nNextFrameTime)
				m_nNextFrameTime = nCurrentTime + m_nTimeStep;
		}
	}

	void DXApplication::handleMouseMove(int x, int y)
	{
		float scaleFactor = 0.005F;
		float yaw = x * scaleFactor;
		float pitch = y * scaleFactor;

		m_pSceneManager->GetCamera()->Rotate(yaw, pitch);
	}

	int DXApplication::Run()
	{
		MSG msg;
		ZeroMemory(&msg, sizeof(MSG));

		while(msg.message != WM_QUIT)
		{
			if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
			else
			{
				makeStep();
			}
		}

		return (int)msg.wParam;
	}

	LRESULT DXApplication::WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		switch(uMsg)
		{
		case WM_SETCURSOR:
			{
				SetCursor(NULL);
				break;
			}

		case WM_INPUT:
			{
				UINT dwSize = 40;
				static BYTE lpb[40];

				GetRawInputData((HRAWINPUT)lParam, RID_INPUT,  lpb, &dwSize, sizeof(RAWINPUTHEADER));

				RAWINPUT* raw = (RAWINPUT*)lpb;

				if(raw->header.dwType == RIM_TYPEMOUSE) 
				{
					int xPosRelative = raw->data.mouse.lLastX;
					int yPosRelative = raw->data.mouse.lLastY;

					handleMouseMove(xPosRelative, yPosRelative);
				}

				break;
			}

		case WM_KEYDOWN:
			{
				m_aKeyboardState[wParam] = true;
				break;
			}

		case WM_KEYUP:
			{
				m_aKeyboardState[wParam] = false;
				break;
			}

		case WM_DESTROY:
			{
				PostQuitMessage(0);
				break;
			}

		default:
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}

		return 0;
	}

	LRESULT CALLBACK DXApplication::WndProcWrapper(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		return Instance()->WndProc(hWnd, uMsg, wParam, lParam);
	}
}
